import pytest
from tests.integration import asserts

from threescale_api_crd.resources import Limits

# With service metric
# tests important for CRD - CRU + list


@pytest.fixture()
def limit_client(application_plan, metric, limit) -> Limits:
    return application_plan.limits(metric)


def test_limit_list(limit_client):
    limits = limit_client.list()
    assert len(limits) >= 1


def test_limit_can_be_created(limit_params, limit, limit_client):
    asserts.assert_resource(limit)
    asserts.assert_resource_params(limit, limit_params)


def test_limit_can_be_read(limit_client, limit, limit_params):
    read = limit_client.read(limit.entity_id)
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, limit_params)


def test_limit_can_be_updated(limit_client, limit):
    lcount = limit_client.list()
    limit["value"] = 11
    lim_updated = limit.update()
    assert limit["value"] == 11 == lim_updated["value"]
    updated = limit.read()
    assert updated["value"] == 11 == limit["value"]
    assert lcount == limit_client.list()


# end of tests important for CRD - CRU + list


# With backend metric
# tests important for CRD - CRU + list


@pytest.fixture()
def backend_limit_client(application_plan, backend_metric, backend_limit) -> Limits:
    return application_plan.limits(backend_metric)


def test_backend_limit_list(backend_limit_client):
    limits = backend_limit_client.list()
    assert len(limits) >= 1


def test_backend_limit_can_be_created(
    backend_limit_params, backend_limit, backend_limit_client
):
    asserts.assert_resource(backend_limit)
    asserts.assert_resource_params(backend_limit, backend_limit_params)


def test_backend_limit_can_be_read(
    backend_limit_client, backend_limit, backend_limit_params
):
    read = backend_limit_client.read(backend_limit.entity_id)
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, backend_limit_params)


def test_backend_limit_can_be_updated(backend_limit_client, backend_limit):
    lcount = backend_limit_client.list()
    backend_limit["value"] = 11
    lim_updated = backend_limit.update()
    assert backend_limit["value"] == 11 == lim_updated["value"]
    updated = backend_limit.read()
    assert updated["value"] == 11 == backend_limit["value"]
    assert lcount == backend_limit_client.list()


# end of tests important for CRD - CRU + list



import json

from tests.integration import asserts
from .asserts import assert_resource, assert_resource_params

# tests important for CRD - CRU + list


def test_activedocs_list(api, active_doc):
    active_docs = api.active_docs.list()
    assert len(active_docs) >= 1


def test_activedoc_can_be_created(api, active_docs_params, active_doc):
    assert_resource(active_doc)
    assert_resource_params(active_doc, active_docs_params)


def test_activedoc_can_be_read(api, active_docs_params, active_doc):
    read = api.active_docs.read(int(active_doc.entity_id))
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, active_docs_params)


def test_activedoc_can_be_read_by_name(api, active_docs_params, active_doc):
    account_name = active_doc["system_name"]
    read = api.active_docs[account_name]
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, active_docs_params)


def test_activedoc_can_be_updated(api, active_doc):
    des_changed = (active_doc["description"] or "") + "_changed"
    active_doc["description"] = des_changed
    gg = active_doc.update()
    assert active_doc["description"] == des_changed
    updated = active_doc.read()
    assert updated["description"] == des_changed
    assert active_doc["description"] == des_changed


# end of tests important for CRD - CRU + list

# tests to compare data in CRD and in 3scale


def test_data(api, api_origin, active_doc):
    for acd in api.active_docs.list():
        acd_origin = api_origin.active_docs.read(int(acd["id"]))
        for att in acd.entity:
            # exclude CRD specific keys
            if att not in ["activeDocOpenAPIRef"]:
                # body should be processed because 3scale processes json
                if att == "body":
                    body = json.loads(acd.entity[att])
                    body_origin = json.loads(acd_origin.entity[att])
                    assert body == body_origin
                else:
                    assert acd.entity[att] == acd_origin.entity[att]



import pytest
import backoff

from threescale_api.errors import ApiClientError

from tests.integration import asserts

# tests important for CRD - CRU + list


def test_should_list_backend_usages(backend, service, backend_usage):
    assert len(service.backend_usages.list()) > 0
    assert len(backend.usages()) > 0


def test_should_create_backend_usage(backend_usage, backend_usage_params):
    asserts.assert_resource(backend_usage)
    asserts.assert_resource_params(backend_usage, backend_usage_params)


def test_should_read_backend_usage(backend_usage, backend_usage_params):
    resource = backend_usage.read()
    asserts.assert_resource(resource)
    asserts.assert_resource_params(resource, backend_usage_params)


def test_should_update_backend_usage(
    service, backend_usage, backend_updated_usage_params
):
    lcount = len(service.backend_usages.list())
    resource = backend_usage.update(params=backend_updated_usage_params)
    asserts.assert_resource(resource)
    asserts.assert_resource_params(resource, backend_updated_usage_params)
    assert lcount == len(service.backend_usages.list())


# end of tests important for CRD - CRU + list



""" Module with default objects """

import logging
import copy
import random
import string
import time
from typing import Dict, List, Union

import threescale_api
import threescale_api.errors
import openshift_client as ocp

LOG = logging.getLogger(__name__)


class DefaultClientCRD(threescale_api.defaults.DefaultClient):
    """Default CRD client."""

    CRD_IMPLEMENTED = False
    SPEC = None
    SELECTOR = None
    KEYS = None
    ID_NAME = None

    def __init__(
        self, parent=None, instance_klass=None, entity_name=None, entity_collection=None
    ):
        ocp.set_default_loglevel(6)
        super().__init__(
            parent=parent,
            instance_klass=instance_klass,
            entity_name=entity_name,
            entity_collection=entity_collection,
        )
        ocp.set_default_project(self.threescale_client.ocp_namespace)

    def get_list(self, typ="normal"):
        """Returns list of entities."""
        return []

    def get_selector(self, obj_name=None):
        """Returns OCP selector for objects."""
        sel = self.SELECTOR + ".capabilities.3scale.net"
        if obj_name:
            sel += "/" + obj_name
        return ocp.selector(sel)

    def read_crd(self, obj_name=None):
        """Read current CRD definition based on selector and/or object name."""
        LOG.info("CRD read %s %s", str(self.SELECTOR), str(obj_name))
        return self.get_selector(obj_name).objects()

    def is_crd_implemented(self):
        """Returns True is crd is implemented in the client"""
        return self.__class__.CRD_IMPLEMENTED

    def enable_crd_implemented(self):
        """Set True to crd is implemented attribute"""
        self.__class__.CRD_IMPLEMENTED = True

    def disable_crd_implemented(self):
        """Set False to crd is implemented attribute"""
        self.__class__.CRD_IMPLEMENTED = False

    def fetch_crd_entity(self, name: str):
        """Fetches the entity based on crd name
        Args:
            name(str): entity crd name
        Returns(obj): Resource
        """
        LOG.info(self._log_message("[FETCH] CRD Fetch by crd name: ", entity_id=name))
        if not self.is_crd_implemented():
            raise threescale_api.errors.ThreeScaleApiError(
                message="Not supported method"
            )
        inst = self._create_instance(response=self.read_crd(name))
        return inst[0] if inst else None

    def read_by_name(self, name: str, **kwargs) -> "DefaultResourceCRD":
        """Read resource by name
        Args:
            name: Name of the resource (either system name, name, org_name ...)
            **kwargs:

        Returns:

        """
        return self.fetch_crd_entity(name) or super().read_by_name(name, **kwargs)

    def read(self, entity_id: int = None, **kwargs) -> 'DefaultResourceCRD':
        """Read the instance, read will just create empty resource and lazyloads only if needed
        Args:
            entity_id(int): Entity id
        Returns(DefaultResourceCRD): Default resource
        """
        LOG.debug(self._log_message("[READ] CRD Read ", entity_id=entity_id))
        if self.is_crd_implemented():
            return self.fetch(entity_id=entity_id, **kwargs)
        else:
            return threescale_api.defaults.DefaultClient.read(self, entity_id, **kwargs)

    def fetch(self, entity_id: int = None, **kwargs):
        """Fetches the entity dictionary
        Args:
            entity_id(int): Entity id
            **kwargs: Optional args

        Returns(dict): Resource dict from the 3scale
        """
        LOG.info(
            self._log_message("[FETCH] CRD Fetch ", entity_id=entity_id, args=kwargs)
        )
        if self.is_crd_implemented():
            list_crds = self.read_crd()
            instance_list = self._create_instance(response=list_crds)
            ret = []
            if isinstance(instance_list, list):
                ret = (
                    [
                        instance
                        for instance in instance_list
                        if (instance.entity_id and instance.entity_id == entity_id)
                        or entity_id is None
                    ][:1]
                    or [None]
                )[0]
            else:
                # proxy.fetch exception
                ret = instance_list
            return ret
        return threescale_api.defaults.DefaultClient.fetch(self, entity_id, **kwargs)

    def exists(self, entity_id=None, **kwargs) -> bool:
        """Check whether the resource exists
        Args:
            entity_id(int): Entity id
            **kwargs: Optional args

        Returns(bool): True if the resource exists
        """
        LOG.info(
            self._log_message(
                "[EXIST] CRD Resource exist ", entity_id=entity_id, args=kwargs
            )
        )
        return self.fetch(entity_id, **kwargs)

    def _list(self, **kwargs) -> List["DefaultResourceCRD"]:
        """Internal list implementation used in list or `select` methods
        Args:
            **kwargs: Optional parameters

        Returns(List['DefaultResourceCRD']):

        """
        LOG.info(self._log_message("[_LIST] CRD", args=kwargs))
        if self.is_crd_implemented():
            list_crds = self.read_crd()
            instance = self._create_instance(response=list_crds, collection=True)
            return instance
        return threescale_api.defaults.DefaultClient._list(self, **kwargs)

    @staticmethod
    def normalize(str_in: str):
        """Some values in CRD cannot contain some characters."""
        return str_in.translate(
            "".maketrans({"-": "", "_": "", "/": "", "[": "", "]": ""})
        ).lower()

    @staticmethod
    def cleanup_spec(spec, keys, params):
        """Removes from spec attributes with None value."""
        for key, value in keys.items():
            if (
                params.get(key, None) is None
                and value in spec["spec"]
                and spec["spec"][value] is None
            ):
                del spec["spec"][value]

    def create(self, params: dict = None, **kwargs) -> "DefaultResourceCRD":
        LOG.info(self._log_message("[CREATE] Create CRD ", body=params, args=kwargs))
        if self.is_crd_implemented():
            spec = copy.deepcopy(self.SPEC)
            name = params.get("name") or params.get(
                "username"
            )  # Developer User exception
            if name is not None:
                name = self.normalize(name)
                if params.get("name"):
                    params["name"] = name
                else:
                    params["username"] = name
            else:
                name = self.normalize(
                    "".join(random.choice(string.ascii_letters) for _ in range(16))
                )

            spec["metadata"]["namespace"] = self.threescale_client.ocp_namespace
            spec["metadata"]["name"] = name
            spec = self._set_provider_ref_new_crd(spec)
            self.before_create(params, spec)

            spec["spec"].update(self.translate_to_crd(params))
            DefaultClientCRD.cleanup_spec(spec, self.KEYS, params)

            result = ocp.create(spec)
            assert result.status() == 0
            # list_objs = self.read_crd(result.out().strip().split('/')[1])
            created_objects = []
            # counters = [89, 55, 34, 21, 13, 8, 5, 3, 2, 1, 1, 1]
            # while len(list_objs) > 0 and len(counters) > 0:
            #    list_objs2 = []
            #    for obj in list_objs:
            #        obj.refresh()
            #        status = obj.as_dict().get('status', None)
            #        if status:
            #            new_id = status.get(self.ID_NAME, 0)
            #            if self._is_ready(status, new_id):
            #                created_objects.append(obj)
            #            else:
            #                list_objs2.append(obj)
            #        else:
            #            list_objs2.append(obj)
            #    list_objs = list_objs2
            #    if not list_objs:
            #        time.sleep(counters.pop())

            timeout = 1000
            # if self.__class__.__name__ in ['Promotes']:
            #    timeout = 1000

            with ocp.timeout(timeout):
                (success, created_objects, _) = result.until_all(
                    success_func=lambda obj: self._is_ready(obj)
                )
                assert created_objects
                assert success

            instance = (self._create_instance(response=created_objects)[:1] or [None])[
                0
            ]
            return instance

        return threescale_api.defaults.DefaultClient.create(self, params, **kwargs)

    def _set_provider_ref_new_crd(self, spec):
        """set provider reference to new crd"""
        if self.threescale_client.ocp_provider_ref is None:
            spec["spec"].pop("providerAccountRef")
        else:
            spec["spec"]["providerAccountRef"][
                "name"
            ] = self.threescale_client.ocp_provider_ref
            spec["spec"]["providerAccountRef"][
                "namespace"
            ] = self.threescale_client.ocp_namespace
        return spec

    def _is_ready(self, obj):
        """Is object ready?"""
        if not ("status" in obj.model and "conditions" in obj.model.status):
            return False
        status = obj.as_dict()["status"]
        new_id = status.get(self.ID_NAME, 0)
        state = {"Failed": True, "Invalid": True, "Synced": False, "Ready": False}
        for sta in status["conditions"]:
            state[sta["type"]] = sta["status"] == "True"

        return (
            not state["Failed"]
            and not state["Invalid"]
            and (state["Synced"] or state["Ready"])
            and (new_id != 0)
        )

    def _create_instance(self, response, klass=None, collection: bool = False):
        klass = klass or self._instance_klass
        if self.is_crd_implemented():
            extracted = self._extract_resource_crd(response, collection, klass=klass)
            instance = self._instantiate_crd(extracted=extracted, klass=klass)
        else:
            extracted = self._extract_resource(response, collection)
            instance = self._instantiate(extracted=extracted, klass=klass)
        LOG.info("[INSTANCE] CRD Created instance: %s", str(instance))
        return instance

    def _extract_resource_crd(self, response, collection, klass):
        extract_params = {"response": response, "entity": self._entity_name}
        if collection:
            extract_params["collection"] = self._entity_collection
        extracted = None
        if isinstance(response, list):
            if response:
                return [{"spec": obj.as_dict()["spec"], "crd": obj} for obj in response]
            return None

        return extracted

    def _instantiate_crd(self, extracted, klass):
        if isinstance(extracted, list):
            instance = [self.__make_instance_crd(item, klass) for item in extracted]
            return self._create_instance_trans(instance)
        if extracted:
            return self.__make_instance_crd(extracted, klass)
        return None

    def _create_instance_trans(self, instance):
        return instance

    def __make_instance_crd(self, extracted: dict, klass):
        instance = (
            klass(client=self, spec=extracted["spec"], crd=extracted["crd"])
            if klass
            else extracted
        )
        return instance

    def delete(
        self, entity_id: int = None, resource: "DefaultResourceCRD" = None, **kwargs
    ) -> bool:
        """Method deletes resource."""
        LOG.info(
            self._log_message("[DELETE] Delete CRD ", entity_id=entity_id, args=kwargs)
        )
        if self.is_crd_implemented():
            resource.crd.delete()
            return True
        return threescale_api.defaults.DefaultClient.delete(
            self, entity_id=entity_id, **kwargs
        )

    def update(
        self,
        entity_id=None,
        params: dict = None,
        resource: "DefaultResourceCRD" = None,
        **kwargs
    ) -> "DefaultResourceCRD":
        LOG.info(
            self._log_message(
                "[UPDATE] Update CRD", body=params, entity_id=entity_id, args=kwargs
            )
        )
        new_params = {}
        if resource:
            new_params = {**resource.entity}
        if params:
            new_params.update(params)
        # TODO change ids for objects which require ids
        if self.is_crd_implemented():
            new_params = copy.deepcopy(new_params)
            self.before_update(new_params, resource)
            new_spec = self.translate_to_crd(new_params)
            if resource.crd is None:
                resource = resource.read()
                if isinstance(resource, list):
                    resource = resource[0]
            else:
                resource.crd.refresh()
            new_crd = resource.crd.as_dict()
            new_crd["spec"].update(new_spec)
            if self.__class__.__name__ not in ["Tenants"]:
                if self.threescale_client.ocp_provider_ref is None:
                    new_crd["spec"].pop("providerAccountRef", None)
                else:
                    new_crd["spec"]["providerAccountRef"] = {
                        "name": self.threescale_client.ocp_provider_ref,
                        # 'namespace': self.threescale_client.ocp_namespace
                    }
            resource.crd.model = ocp.Model(new_crd)
            result = resource.crd.replace()

            if result.status():
                LOG.error("[INSTANCE] Update CRD failed: %s", str(result))
                raise Exception(str(result))
            # return self.read(resource.entity_id)
            return resource

        return threescale_api.defaults.DefaultClient.update(
            self, entity_id=entity_id, params=params, **kwargs
        )

    def trans_item(self, key, value, obj):
        """Transform one attribute in CRD spec."""
        return obj[key]

    def translate_to_crd(self, obj):
        """Translate object attributes into object ready for merging into CRD."""
        map_ret = {}

        for key, value in self.KEYS.items():
            LOG.debug("%s, %s, %s, %s", str(key), str(value), str(obj), str(type(obj)))
            if obj.get(key, None) is not None:
                set_value = self.trans_item(key, value, obj)
                if set_value is not None:
                    map_ret[value] = set_value

        return map_ret


class DefaultClientNestedCRD(DefaultClientCRD):
    """Default CRD client for nested objects."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_id_from_crd(self):
        """Returns object id extracted from CRD."""
        return None

    # flake8: noqa C901
    def _extract_resource_crd(self, response, collection, klass) -> Union[List, Dict]:
        extract_params = {"response": response, "entity": self._entity_name}
        if collection:
            extract_params["collection"] = self._entity_collection
        extracted = None
        if isinstance(response, list):
            if self.is_crd_implemented():
                parent_id = int(self.topmost_parent().entity_id)
                service_with_maps = {}
                for prod in response:
                    prod_dict = prod.as_dict()
                    if prod_dict is None:
                        prod_dict = {}
                    idp = int(
                        prod_dict.get("status", {}).get(
                            self.topmost_parent().client.ID_NAME, 0
                        )
                    )
                    if idp == parent_id:
                        service_with_maps = prod
                        break
                spec = {}
                if service_with_maps != {}:
                    spec = (
                        DictQuery(service_with_maps.as_dict()).get(
                            klass.GET_PATH or self.get_path()
                        )
                    ) or []
                if isinstance(spec, list):
                    return [{"spec": obj, "crd": service_with_maps} for obj in spec]
                elif (
                    "apicastHosted" not in spec.keys()
                ):  # dict branch, exception for Proxy
                    ret = []
                    for key, obj in spec.items():
                        obj[self.ID_NAME] = key

                        ret.append({"spec": obj, "crd": service_with_maps})
                    return ret
                else:
                    return [{"spec": spec, "crd": service_with_maps}]

        return extracted

    def create(self, params: dict = None, **kwargs) -> "DefaultResourceCRD":
        LOG.info(
            self._log_message("[CREATE] Create CRD Nested ", body=params, args=kwargs)
        )
        if self.is_crd_implemented():
            spec = copy.deepcopy(self.SPEC)
            name = params.get("name") or params.get(
                "username"
            )  # Developer User exception
            if name is not None:
                name = self.normalize(name)
                if params.get("name"):
                    params["name"] = name
                else:
                    params["username"] = name
            else:
                params["name"] = self.normalize(
                    "".join(random.choice(string.ascii_letters) for _ in range(16))
                )
                name = params["name"]

            self.before_create(params, spec)

            spec["spec"].update(self.translate_to_crd(params))
            DefaultClientCRD.cleanup_spec(spec, self.KEYS, params)

            return self.in_create(self.get_list_from_spec(), params, spec)

        return threescale_api.defaults.DefaultClient.create(self, params, **kwargs)

    def delete(
        self, entity_id: int = None, resource: "DefaultResourceCRD" = None, **kwargs
    ) -> bool:
        """Method deletes resource."""
        LOG.info(
            self._log_message(
                "[DELETE] Delete CRD Nested ", entity_id=entity_id, args=kwargs
            )
        )
        if self.is_crd_implemented():
            spec = self.translate_to_crd(resource.entity)
            maps = self.remove_from_list(spec)
            self.update_list(maps)
            return True
        return threescale_api.defaults.DefaultClient.delete(
            self, entity_id=entity_id, **kwargs
        )

    def update(
        self,
        entity_id=None,
        params: dict = None,
        resource: "DefaultResourceCRD" = None,
        **kwargs
    ) -> "DefaultResourceCRD":
        LOG.info(
            self._log_message(
                "[UPDATE] Update CRD", body=params, entity_id=entity_id, args=kwargs
            )
        )
        new_params = {}
        if resource:
            new_params = {**resource.entity}
        if params:
            new_params.update(params)
        # TODO change ids for objects which require ids
        if self.is_crd_implemented():
            # ApplicationPlans
            # BackendMappingRules
            # BackendMetrics
            # BackendUsages
            # Limits
            # MappingRules
            # Metrics
            # Policies
            # PricingRules
            # Proxies
            # if self.__class__.__name__ == 'BackendUsages':
            spec = self.before_update(new_params, resource)

            maps = self.remove_from_list(spec)

            # par = self.parent
            maps = self.before_update_list(maps, new_params, spec, resource)

            par = self.update_list(maps)
            maps = self.get_list()
            return self.after_update_list(maps, par, new_params)

        return threescale_api.defaults.DefaultClient.update(
            self, entity_id=entity_id, params=params, **kwargs
        )

    def after_update_list(self, maps, par, new_params):
        """Returns updated list."""
        checked_keys = [key for key in new_params.keys() if key not in ["id"]]
        for mapi in maps:
            if all([new_params[key] == mapi[key] for key in checked_keys]):
                return mapi
        return None


class DefaultResourceCRD(threescale_api.defaults.DefaultResource):
    """Default CRD resource."""

    GET_PATH = None

    def __init__(self, *args, crd=None, **kwargs):
        super().__init__(**kwargs)
        self._crd = crd

    @property
    def crd(self):
        """CRD object property."""
        if not self._crd:
            self.read()
        return self._crd or self.entity.get("crd", None)

    @crd.setter
    def crd(self, value):
        self._crd = value

    @property
    def entity_id(self) -> int:
        return self._entity_id or self._entity.get("id") or self.get_id_from_crd()

    @entity_id.setter
    def entity_id(self, value):
        self._entity_id = value

    def get_id_from_crd(self):
        """Returns object id extracted from CRD."""
        counter = 5
        while counter > 0:
            self.crd = self.crd.refresh()
            status = self.crd.as_dict()["status"]
            ret_id = status.get(self.client.ID_NAME, None)
            if ret_id:
                return ret_id
            time.sleep(20)
            counter -= 1

        return None

    def get_path(self):
        """
        This function is usefull only for Limits and PricingRules and\
        it should be redefined there.
        """
        return self.__class__.GET_PATH


class DictQuery(dict):
    """Get value from nested dictionary."""

    def get(self, path, default=None):
        keys = path.split("/")
        val = None

        for key in keys:
            if val:
                if isinstance(val, list):
                    val = [v.get(key, default) if v else None for v in list(val)]
                else:
                    val = val.get(key, default)
            else:
                val = dict.get(self, key, default)

            if not val:
                break

        return val



import secrets
import random
import string
import pytest

from tests.integration import asserts


# tests important for CRD - CRU + list
def test_application_list_global(api, application):
    applications = api.applications.list()
    assert len(applications) >= 1


def test_application_list(account, application):
    applications = account.applications.list()
    assert len(applications) >= 1


def test_application_can_be_created(application_params, application):
    asserts.assert_resource(application)
    asserts.assert_resource_params(application, application_params)


def test_application_can_be_read(api, application_params, application):
    read = api.applications.read(int(application.entity_id))
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, application_params)


def test_application_can_be_read_by_name(api, application_params, application):
    app_name = application["name"]
    read = api.applications[app_name]
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, application_params)


@pytest.fixture(scope="module")
def application_plan_params2():
    suffix = secrets.token_urlsafe(8)
    return {
        "name": f"test-{suffix}",
        "setup_fee": "1.00",
        "state_event": "publish",
        "cost_per_month": "3.00",
    }


@pytest.fixture(scope="module")
def application_plan2(service, application_plan_params2):
    resource = service.app_plans.create(params=application_plan_params2)
    yield resource
    resource.delete()


@pytest.fixture(scope="module")
def update_application_params(application_plan2):
    suffix = secrets.token_urlsafe(8)
    name = f"updated-{suffix}"
    return {"name": name, "description": name, "plan_id": application_plan2["id"]}


def test_application_update(update_application_params, application):
    updated_application = application.update(params=update_application_params)
    asserts.assert_resource(updated_application)
    asserts.assert_resource_params(updated_application, update_application_params)


# tests important for CRD - CRU + list

# changing application state tests


def test_application_set_state(application):
    application = application.set_state("suspend")
    assert application["state"] == "suspended"
    application = application.set_state("resume")
    assert application["state"] == "live"


# Application Atuhentication - Application keys


def test_application_key_list(application, app_key):
    keys = application.keys.list()
    assert len(keys) > 0


def test_application_key_can_be_created(app_key, app_key_params):
    asserts.assert_resource(app_key)
    asserts.assert_resource_params(app_key, app_key_params)


def test_application_autogenerated_key_can_be_created(application, app_key_params):
    keys_len = len(application.keys.list())
    key_params = app_key_params.copy()
    key_params.pop("key", None)
    key_params["generateSecret"] = True
    new_key = application.keys.create(params=key_params)
    asserts.assert_resource(new_key)
    asserts.assert_resource_params(new_key, key_params)
    assert new_key["value"]
    assert len(application.keys.list()) == keys_len + 1


def test_application_update_userkey(application):
    new_key = "".join(
        random.choices(string.ascii_letters + string.digits + "-_.", k=100)
    )
    application.update(params={"user_key": new_key})
    application.read()
    asserts.assert_resource(application)
    assert application["user_key"] == new_key



from tests.integration import asserts
from .asserts import assert_resource, assert_resource_params


def test_3scale_master_url_is_set(master_api, master_url, master_token):
    assert master_url
    assert master_token
    assert master_api.url


# tests important for CRD - CRU + list


def test_tenants_list(api, custom_tenant):
    tenants = api.tenants.list()
    assert len(tenants) >= 1


def test_tenant_can_be_created(custom_tenant, tenant_params):
    asserts.assert_resource(custom_tenant)
    assert custom_tenant.entity["signup"]["account"]["admin_base_url"]
    assert custom_tenant.entity["signup"]["access_token"]["value"]


def test_tenant_can_be_read(api, tenant_params, custom_tenant):
    read = api.tenants.read(custom_tenant.entity_id)
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, tenant_params)


# this test does not make sense in 3scale api client too
#   def test_tenant_can_be_read_by_name(api, tenant_params, custom_tenant):
#       tenant_name = custom_tenant['name']
#       read = api.tenants[tenant_name]
#       asserts.assert_resource(read)
#       asserts.assert_resource_params(read, tenant_params)


def test_tenant_can_be_updated(api, custom_tenant):
    des_changed = (custom_tenant["signup"]["account"]["email"] or "") + "_changed"
    custom_tenant["signup"]["account"]["email"] = des_changed
    custom_tenant.update()
    assert custom_tenant["signup"]["account"]["email"] == des_changed
    updated = custom_tenant.threescale_client.tenants.read(custom_tenant.entity_id)
    assert updated["signup"]["account"]["email"] == des_changed
    assert custom_tenant["signup"]["account"]["email"] == des_changed


# end of tests important for CRD - CRU + list


def test_tenant_annotation(custom_tenant, master_api_origin):
    tenant_origin = master_api_origin.tenants.read(custom_tenant.entity_id)
    assert tenant_origin["signup"]["account"]["annotations"]["managed_by"] == "operator"



import re
import sys

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = "devel"
if sys.argv[1] == "--release-version":
    sys.argv.pop(1)
    VERSION = sys.argv.pop(1)
    assert re.match(
        r"[0-9]+\.[0-9]+\.[0-9]+", VERSION
    ), "Version definition required as first arg"

requirements = ["3scale-api", "openshift-client"]

extra_requirements = {
    "dev": ["flake8", "mypy", "pylint", "pytest", "python-dotenv", "backoff"],
    "docs": ["sphinx"],
}

setup(
    name="3scale-api-crd",
    version=VERSION,
    description="3scale CRD Python Client",
    author="Martin Kudlej",
    author_email="kudlej.martin@gmail.com",
    maintainer="Matej Dujava",
    maintainer_email="mdujava@redhat.com",
    url="https://github.com/3scale-qe/3scale-api-python-crd",
    packages=find_packages(exclude=("tests",)),
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=requirements,
    extras_require=extra_requirements,
    entry_points={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
    ],
)



import os
import secrets
import random
import string

import pytest
from dotenv import load_dotenv

# from threescale_api import errors

import threescale_api
import threescale_api_crd

from threescale_api_crd.resources import (
    Service,
    Metric,
    BackendUsage,
    BackendMappingRule,
    Backend,
    ActiveDoc,
    PolicyRegistry,
    MappingRule,
    Limit,
    ApplicationPlan,
    Proxy,
    PricingRule,
    Application,
)

load_dotenv()

def strtobool(istr):
    return istr == "y" or istr == "yes" or istr == "t" or istr == "true" or istr == "1"


def cleanup(resource):
    resource.delete()
    assert not resource.exists()


def get_suffix() -> str:
    return secrets.token_urlsafe(8)


@pytest.fixture(scope="session")
def url() -> str:
    return os.getenv("THREESCALE_PROVIDER_URL")


@pytest.fixture(scope="session")
def token() -> str:
    return os.getenv("THREESCALE_PROVIDER_TOKEN")


@pytest.fixture(scope="session")
def master_url() -> str:
    return os.getenv("THREESCALE_MASTER_URL")


@pytest.fixture(scope="session")
def master_token() -> str:
    return os.getenv("THREESCALE_MASTER_TOKEN")


@pytest.fixture(scope="session")
def ssl_verify() -> bool:
    ssl_verify = os.getenv("THREESCALE_SSL_VERIFY", "false")
    ssl_verify = strtobool(ssl_verify)
    if not ssl_verify:
        import urllib3

        urllib3.disable_warnings()
    return ssl_verify


@pytest.fixture(scope="session")
def api_backend() -> str:
    return os.getenv("TEST_API_BACKEND", "http://www.httpbin.org:80")


@pytest.fixture(scope="session")
def ocp_provider_ref() -> str:
    return os.getenv("OCP_PROVIDER_ACCOUNT_REF")


@pytest.fixture(scope="session")
def api(
    url: str, token: str, ssl_verify: bool, ocp_provider_ref: str
) -> threescale_api_crd.ThreeScaleClientCRD:
    return threescale_api_crd.ThreeScaleClientCRD(
        url=url, token=token, ssl_verify=ssl_verify, ocp_provider_ref=ocp_provider_ref
    )


@pytest.fixture(scope="session")
def api_origin(
    url: str, token: str, ssl_verify: bool) -> threescale_api.ThreeScaleClient:
    return threescale_api.ThreeScaleClient(url=url, token=token, ssl_verify=ssl_verify)


@pytest.fixture(scope="session")
def master_api(
    master_url: str, master_token: str, ssl_verify: bool, ocp_provider_ref: str
) -> threescale_api_crd.ThreeScaleClientCRD:
    return threescale_api_crd.ThreeScaleClientCRD(
        url=master_url,
        token=master_token,
        ssl_verify=ssl_verify,
        ocp_provider_ref=ocp_provider_ref,
    )


@pytest.fixture(scope="session")
def master_api_origin(
    master_url: str, master_token: str, ssl_verify: bool) -> threescale_api.ThreeScaleClient:
    return threescale_api.ThreeScaleClient(
        url=master_url,
        token=master_token,
        ssl_verify=ssl_verify,
    )


@pytest.fixture(scope="module")
def apicast_http_client(application, proxy, ssl_verify):
    proxy.deploy()
    return application.api_client(verify=ssl_verify)


@pytest.fixture(scope="module")
def service_params():
    suffix = get_suffix()
    return {"name": f"test-{suffix}"}


@pytest.fixture(scope="module")
def service(service_params, api) -> Service:
    service = api.services.create(params=service_params)
    yield service
    cleanup(service)


@pytest.fixture(scope="module")
def account_params():
    suffix = get_suffix()
    name = f"testacc{suffix}"
    return {
        "name": name,
        "username": name,
        "org_name": name,
        "monthly_billing_enabled": False,
        "monthly_charging_enabled": False,
        "email": f"{name}@name.none",
    }


@pytest.fixture(scope="module")
def account(api, account_params):
    acc = api.accounts.create(params=account_params)
    account_params.update(account_name=acc["name"])
    yield acc
    cleanup(acc)


@pytest.fixture(scope="module")
def acc_user(account):
    return account.users.list()[-1]


@pytest.fixture(scope="module")
def acc_user2_params(account, acc_user):
    name = acc_user["username"] + "2"
    return {
        "username": name,
        "email": f"{name}@name.none",
        "role": "member",
        "account_name": account["name"],
    }


@pytest.fixture(scope="module")
def acc_user2(account, acc_user, acc_user2_params):
    return account.users.create(acc_user2_params)


@pytest.fixture(scope="module")
def application_plan_params(service) -> dict:
    suffix = get_suffix()
    return {
        "name": f"test-{suffix}",
        "setup_fee": "1.00",
        "state_event": "publish",
        "cost_per_month": "3.00",
    }


@pytest.fixture(scope="module")
def application_plan(api, service, application_plan_params) -> ApplicationPlan:
    resource = service.app_plans.create(params=application_plan_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def application_params(application_plan, service, account):
    suffix = get_suffix()
    name = f"test-{suffix}"
    return {
        "name": name,
        "description": name,
        "plan_id": application_plan["id"],
        "service_id": service["id"],
        "account_id": account["id"],
    }


@pytest.fixture(scope="module")
def application(service, application_plan, application_params, account) -> Application:
    resource = account.applications.create(params=application_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def app_key_params(account, application):
    value = "".join(
        random.choices(
            string.ascii_uppercase
            + string.digits
            + "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~",
            k=100,
        )
    )
    return {
        "application_id": application["id"],
        "account_id": account["id"],
        "key": value,
    }


@pytest.fixture(scope="module")
def app_key(application, app_key_params) -> threescale_api.resources.ApplicationKey:
    resource = application.keys.create(params=app_key_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def proxy(service) -> Proxy:
    return service.proxy.list()


@pytest.fixture(scope="module")
def backend_usage_params(service, backend):
    return {
        "service_id": service["id"],
        "backend_id": backend["id"],
        "path": "/",
    }


@pytest.fixture(scope="module")
def backend_updated_usage_params(backend_usage_params):
    ret = backend_usage_params.copy()
    ret["path"] = "/post"
    return ret


@pytest.fixture(scope="module")
def backend_usage(service, backend, backend_usage_params) -> BackendUsage:
    service = service.read()
    resource = service.backend_usages.create(params=backend_usage_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def metric_params(service):
    suffix = get_suffix()
    friendly_name = f"test-metric-{suffix}"
    name = f"{friendly_name}".replace("-", "_")
    return {"friendly_name": friendly_name, "system_name": name, "unit": "count"}


@pytest.fixture(scope="module")
def backend_metric_params():
    suffix = get_suffix()
    friendly_name = f"test-metric-{suffix}"
    name = f"{friendly_name}".replace("-", "")
    return {"friendly_name": friendly_name, "system_name": name, "unit": "count"}


@pytest.fixture(scope="module")
def updated_metric_params(metric_params):
    suffix = get_suffix()
    friendly_name = f"test-updated-metric-{suffix}"
    metric_params["friendly_name"] = f"/get/{friendly_name}"
    return metric_params


@pytest.fixture(scope="module")
def backend_updated_metric_params(backend_metric_params):
    updated = backend_metric_params.copy()
    suffix = get_suffix()
    friendly_name = f"test-updated-metric-{suffix}"
    updated["friendly_name"] = f"/get/{friendly_name}"
    return updated


@pytest.fixture(scope="module")
def metric(service, metric_params) -> Metric:
    resource = service.metrics.create(params=metric_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def method_params():
    suffix = get_suffix()
    friendly_name = f"test-method-{suffix}"
    system_name = f"{friendly_name}".replace("-", "_")
    return {"friendly_name": friendly_name, "system_name": system_name}


# 'friendly_name' is id in CRD for methods
@pytest.fixture(scope="module")
def updated_method_params(method_params):
    suffix = get_suffix()
    description = f"test-updated-method-{suffix}"
    method_params["description"] = description
    return method_params


@pytest.fixture(scope="module")
def method(hits_metric, method_params):
    resource = hits_metric.methods.create(params=method_params)
    yield resource
    cleanup(resource)
    # service.proxy.deploy()


@pytest.fixture(scope="module")
def hits_metric(service):
    return service.metrics.read_by(system_name="hits")


def get_mapping_rule_pattern():
    suffix = get_suffix()
    pattern = f"test-{suffix}".replace("_", "-")
    return pattern


@pytest.fixture(scope="module")
def mapping_rule_params(service):
    """
    Fixture for getting paramteres for mapping rule for product/service.
    """
    hits_metric = service.metrics.read_by_name("hits")
    return {
        "http_method": "GET",
        "pattern": "/get",
        "metric_id": hits_metric["id"],
        "delta": 1,
    }


@pytest.fixture(scope="module")
def backend_mapping_rule_params(backend, backend_metric):
    """
    Fixture for getting paramteres for mapping rule for backend.
    """
    back = backend_metric["id"]
    return {
        "http_method": "GET",
        "pattern": "/anything/get/ida",
        "metric_id": back,
        "delta": 1,
    }


@pytest.fixture(scope="module")
def updated_mapping_rules_params(mapping_rule_params):
    """
    Fixture for updating mapping rule for product/service.
    """
    pattern = get_mapping_rule_pattern()
    params = mapping_rule_params.copy()
    params["pattern"] = f"/anything/get/{pattern}"
    return params


@pytest.fixture(scope="module")
def updated_backend_mapping_rules_params(backend_mapping_rule_params):
    """
    Fixture for updating mapping rule for backend.
    """
    pattern = get_mapping_rule_pattern()
    params = backend_mapping_rule_params.copy()
    params["pattern"] = f"/anything/get/{pattern}"
    return params


@pytest.fixture(scope="module")
def mapping_rule(service, mapping_rule_params, proxy) -> MappingRule:
    """
    Fixture for getting mapping rule for product/service.
    """
    resource = service.mapping_rules.create(params=mapping_rule_params)
    yield resource
    cleanup(resource)
    proxy.deploy()


@pytest.fixture(scope="module")
def backend_mapping_rule(
    backend, backend_metric, backend_mapping_rule_params, service, proxy, backend_usage
) -> BackendMappingRule:
    """
    Fixture for getting mapping rule for backend.
    """
    backend = backend.read()
    resource = backend.mapping_rules.create(params=backend_mapping_rule_params)
    yield resource
    cleanup(resource)
    proxy.deploy()


@pytest.fixture(scope="module")
def create_mapping_rule(service):
    """
    Fixture for creating mapping rule for product/service.
    """
    rules = []

    def _create(metric, http_method, path):
        params = {
            "service_id": service["id"],
            "http_method": http_method,
            "pattern": f"/anything{path}",
            "delta": 1,
            "metric_id": metric["id"],
        }
        rule = service.mapping_rules.create(params=params)
        rules.append(rule)
        return rule

    yield _create

    # TODO
    # for rule in rules:
    #     if rule.exists():
    #         cleanup(rule)


@pytest.fixture(scope="module")
def create_backend_mapping_rule(backend):
    """
    Fixture for creating mapping rule for backend.
    """
    rules = []

    def _create(backend_metric, http_method, path):
        params = {
            "backend_id": backend["id"],
            "http_method": http_method,
            "pattern": f"/anything{path}",
            "delta": 1,
            "metric_id": backend_metric["id"],
        }
        rule = backend.mapping_rules.create(params=params)
        rules.append(rule)
        return rule

    yield _create

    for rule in rules:
        if rule.exists():
            cleanup(rule)


@pytest.fixture(scope="module")
def backend_params(api_backend):
    """
    Fixture for getting backend parameters.
    """
    suffix = get_suffix()
    return {
        "name": f"test-backend-{suffix}",
        "private_endpoint": api_backend,
        "description": "111",
    }


@pytest.fixture(scope="module")
def backend(backend_params, api) -> Backend:
    """
    Fixture for getting backend.
    """
    backend = api.backends.create(params=backend_params)
    yield backend
    cleanup(backend)


@pytest.fixture(scope="module")
def backend_metric(backend, backend_metric_params) -> Metric:
    """
    Fixture for getting backend metric.
    """
    resource = backend.metrics.create(params=backend_metric_params)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def custom_tenant(master_api, tenant_params):
    """
    Fixture for getting the custom tenant.
    """
    resource = master_api.tenants.create(tenant_params)
    yield resource
    resource.delete()


@pytest.fixture(scope="module")
def tenant_params():
    """
    Params for custom tenant
    """
    return {
        "username": f"tenant{get_suffix()}",
        "admin_password": "123456",
        "email": f"e{get_suffix()}@invalid.invalid",
        "org_name": "org",
    }


@pytest.fixture(scope="module")
def active_docs_body():
    return """{"openapi":"3.0.0","info":{"version":"1.0.0","title":"example"},"paths":{
            "/pets":{"get":{"summary":"List all pets","operationId":"listPets","responses":
            {"200":{"description":"A paged array of pets"}}}}},"components":{}}"""


@pytest.fixture(scope="module")
def active_docs_params(active_docs_body):
    suffix = get_suffix()
    name = f"test-{suffix}"
    des = f"description-{suffix}"
    return {"name": name, "body": active_docs_body, "description": des}


@pytest.fixture(scope="module")
def active_doc(api, service, active_docs_params) -> ActiveDoc:
    """
    Fixture for getting active doc.
    """
    acs = active_docs_params.copy()
    acs["service_id"] = service["id"]
    resource = api.active_docs.create(params=acs)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def openapi_params(active_docs_body):
    suffix = get_suffix()
    name = f"test-{suffix}"
    params = {
        "name": name,
        "productionPublicBaseURL": "http://productionPublicBaseURL",
        "stagingPublicBaseURL": "http://stagingPublicBaseURL",
        "productSystemName": "PrOdUcTsYsTeMnAmE",
        "privateBaseURL": "http://privateBaseURL",
        "prefixMatching": True,
        "privateAPIHostHeader": "privateAPIHostHeader",
        "privateAPISecretToken": "privateAPISecretToken",
        "body": active_docs_body,
    }
    return params


@pytest.fixture(scope="module")
def openapi(api, openapi_params):
    """
    Fixture for getting OpenApi.
    """
    resource = api.openapis.create(params=openapi_params.copy())
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def webhook(api):
    return api.webhooks


@pytest.fixture(scope="session")
def policy_registry_schema():
    return {
        "summary": "This is just an example.",
        "description": [
            "This policy is just an example\
            how to write your custom policy.\nAnd this is next line",
            "And next item.",
        ],
        "name": "APIcast Example Policy",
        "$schema": "http://apicast.io/policy-v1/schema#manifest#",
        "version": "0.1",
        "configuration": {
            "properties": {
                "property1": {
                    "description": "list of properties1",
                    "items": {
                        "properties": {
                            "value1": {"description": "Value1", "type": "string"},
                            "value2": {"description": "Value2", "type": "string"},
                        },
                        "required": ["value1"],
                        "type": "object",
                    },
                    "type": "array",
                }
            },
            "type": "object",
        },
    }


@pytest.fixture(scope="module")
def policy_registry_params(policy_registry_schema):
    """Params for policy registry."""
    suffix = get_suffix()
    name = f"test-{suffix}"
    return {"name": name, "version": "0.1", "schema": policy_registry_schema}


@pytest.fixture(scope="module")
def policy_registry(api, policy_registry_params) -> PolicyRegistry:
    """
    Fixture for getting policy registry.
    """
    acs = policy_registry_params.copy()
    resource = api.policy_registry.create(params=acs)
    yield resource
    cleanup(resource)


@pytest.fixture(scope="module")
def limit_params(metric):
    """Params for limit."""
    return {"metric_id": metric["id"], "period": "minute", "value": 10}


@pytest.fixture(scope="module")
def limit(service, application, application_plan, metric, limit_params) -> Limit:
    """
    Fixture for getting limit.
    """
    resource = application_plan.limits(metric).create(limit_params)
    yield resource
    resource.delete()


@pytest.fixture(scope="module")
def backend_limit_params(backend_metric):
    """Params for limit."""
    return {"metric_id": backend_metric["id"], "period": "minute", "value": 10}


@pytest.fixture(scope="module")
def backend_limit(
    service,
    application,
    application_plan,
    backend_metric,
    backend_limit_params,
    backend_usage,
) -> Limit:
    """
    Fixture for getting limit for backend metric.
    """
    resource = application_plan.limits(backend_metric).create(backend_limit_params)
    yield resource
    resource.delete()


@pytest.fixture(scope="module")
def prule_params(metric):
    """Params for prule."""
    return {"metric_id": metric["id"], "min": 10, "max": 100, "cost_per_unit": "10"}


@pytest.fixture(scope="module")
def prule(service, application, application_plan, metric, prule_params) -> PricingRule:
    """
    Fixture for getting prule.
    """
    resource = application_plan.pricing_rules(metric).create(prule_params)
    yield resource
    resource.delete()


@pytest.fixture(scope="module")
def backend_prule_params(backend_metric):
    """Params for prule."""
    return {
        "metric_id": backend_metric["id"],
        "min": 10,
        "max": 100,
        "cost_per_unit": 10,
    }


@pytest.fixture(scope="module")
def backend_prule(
    service,
    application,
    application_plan,
    backend_metric,
    backend_prule_params,
    backend_usage,
) -> PricingRule:
    """
    Fixture for getting prule for backend metric.
    """
    resource = application_plan.pricing_rules(backend_metric).create(
        backend_prule_params
    )
    yield resource
    resource.delete()


@pytest.fixture(scope="module")
def promote_params(api, service):
    """
    Promote params for service.
    """
    return {"productCRName": service.crd.as_dict()["metadata"]["name"]}


@pytest.fixture(scope="module")
def promote(api, service, backend, backend_usage, promote_params):
    """
    Promote object for service.
    """
    resource = api.promotes.create(params=promote_params)
    yield resource
    cleanup(resource)



import json

from tests.integration import asserts
from .asserts import assert_resource, assert_resource_params

# tests important for CRD - CRU + list


def test_openapis_list(api, openapi):
    openapis = api.openapis.list()
    assert len(openapis) >= 1


def test_openapi_can_be_created(api, openapi_params, openapi):
    assert_resource(openapi)
    assert_resource_params(openapi, openapi_params)


def test_openapi_can_be_read(api, openapi_params, openapi):
    read = api.openapis.read(openapi.entity_id)
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, openapi_params)


def test_openapi_can_be_updated(api, openapi):
    des_changed = (openapi["productSystemName"] or "") + "_changed"
    openapi["productSystemName"] = des_changed
    openapi.update()
    assert openapi["productSystemName"] == des_changed
    updated = openapi.read()
    assert updated["productSystemName"] == des_changed
    assert openapi["productSystemName"] == des_changed


# end of tests important for CRD - CRU + list


def test_openapi_annotation(openapi, api_origin):
    service_origin = api_origin.services.fetch(openapi.service.entity_id)
    assert service_origin['annotations']['managed_by'] == 'operator'

    for backend in openapi.backends:
        backend_origin = api_origin.backends.fetch(backend.entity_id)
        assert backend_origin['annotations']['managed_by'] == 'operator'
